@@ -13,7 +13,7 @@ from typing import Optional, Dict, Callable, Union, List
 
 from transformation_algebra import error
 from transformation_algebra.type import \
-    Type, TypeVar, TypeSchema, TypeInstance, Function, _
+    Type, TypeVar, TypeSchema, TypeInstance, Function
 
 
 class Definition(ABC):
@@ -67,30 +67,43 @@ class Operation(Definition):
         super().__init__(*nargs, **kwargs)
         assert self.type.is_function()
 
+    def validate(self) -> None:
+        """
+        This method raises an error if the operation is a composite operation,
+        but the declared type cannot be reconciled with the type inferred from
+        the composition.
+        """
         # If the operation is composite, check that its declared type is no
         # more general than the type we can infer from the composition function
-        if self.composition:
-            mock = Definition(_)
-            args = [Base(mock) for a in signature(self.composition).parameters]
-            output = self.composition(*args)
-            if isinstance(output, Expr):
-                type_infer = output.type
-                type_decl = self.type.instance()
-                vars_decl = list(type_decl.variables())
-                for a in reversed(args):
-                    type_infer = Function(a.type, type_infer)
-                type_decl.unify(type_infer, subtype=True)
-                type_decl = type_decl.resolve()
+        try:
+            if self.composition:
+                mock = Definition(TypeSchema(lambda: TypeVar()))
+                args = [
+                    Base(mock) for a in signature(self.composition).parameters]
+                output = self.composition(*args)
+                if isinstance(output, Expr):
+                    type_infer = output.type
+                    type_decl = self.type.instance()
+                    vars_decl = list(type_decl.variables())
+                    for a in reversed(args):
+                        type_infer = Function(a.type, type_infer)
+                    type_decl.unify(type_infer, subtype=True)
+                    type_decl = type_decl.resolve()
 
-                # All the variables in the declared type must still be
-                # variables --- otherwise we were too general
-                if not all(isinstance(v.follow(), TypeVar) for v in vars_decl):
-                    raise error.DeclaredTypeTooGeneral(self.type, type_infer)
+                    # All the variables in the declared type must still be
+                    # variables --- otherwise we were too general
+                    if not all(isinstance(v.follow(), TypeVar)
+                            for v in vars_decl):
+                        raise error.DeclaredTypeTooGeneral(
+                            self.type, type_infer)
 
-            else:
-                # The type could not be derived because the result is not a
-                # full expression. Shouldn't happen?
-                raise error.PartialPrimitive()
+                else:
+                    # The type could not be derived because the result is not a
+                    # full expression. Shouldn't happen?
+                    raise error.PartialPrimitive()
+        except error.TAError as e:
+            e.definition = self
+            raise
 
 
 class PartialExpr(ABC):
@@ -137,11 +150,11 @@ class Expr(PartialExpr):
         Obtain a tree representation using Unicode block characters.
         """
         if isinstance(self, Base):
-            return f"╼ {self} : {self.definition.type}"
+            return f"╼ {self} : {self.type.str_with_constraints()}"
         else:
             assert isinstance(self, Application)
             return (
-                f"{self.type.strc()}\n"
+                f"{self.type.str_with_constraints()}\n"
                 f"{lvl} ├─{self.f.tree(lvl + ' │ ')}\n"
                 f"{lvl} └─{self.x.tree(lvl + '   ')}"
             )
@@ -291,9 +304,16 @@ class TransformationAlgebra(object):
     def __setitem__(self, key: str, value: Definition) -> None:
         self.definitions[key.lower()] = value
 
+        # Validation only happens once an operation is added to the algebra. If
+        # we did it at define-time, it would lead to issues --- see issue #3
+        if isinstance(value, Operation):
+            value.validate()
+
     def parse(self, string: str) -> Optional[Expr]:
         # This used to be done via pyparsing, but the structure is so simple
         # that I opted to remove the dependency --- this is *much* faster
+
+        labels: Dict[str, TypeInstance] = dict()
         stack: List[Optional[Expr]] = [None]
 
         for token_group, chars in groupby(string, Token.ize):
@@ -324,6 +344,10 @@ class TransformationAlgebra(object):
                 if previous and isinstance(previous, Base) \
                         and isinstance(previous.definition, Data):
                     previous.label = token
+                    if token in labels:
+                        labels[token].unify(previous.type)
+                    else:
+                        labels[token] = previous.type
                     stack.append(previous)
                 else:
                     try:
