@@ -57,7 +57,7 @@ class Type(ABC):
         # don't have to do anything here except perform a sanity check.
         t = self.instance()
         constraint.set_context(t)
-        for v in constraint.variables():
+        for v in constraint.variables_iter():
             if not v.wildcard and v not in t:
                 raise error.ConstrainFreeVariable(constraint)
         return t
@@ -80,10 +80,12 @@ class Type(ABC):
         return self != other and self >= other
 
     def __le__(self, other: Type) -> Optional[bool]:
-        return self.instance().subtype(other.instance())
+        return self.instance().unifiable(other.instance(),
+            accept_wildcard=True, subtype=True)
 
     def __ge__(self, other: Type) -> Optional[bool]:
-        return self.instance().subtype(other.instance())
+        return self.instance().unifiable(other.instance(),
+            accept_wildcard=True, subtype=True)
 
     def apply(self, arg: Type) -> TypeInstance:
         """
@@ -143,7 +145,7 @@ class TypeSchema(Type):
     def __str__(self) -> str:
         return self.schema(
             *(TypeVar(v) for v in signature(self.schema).parameters)
-        ).resolve().strc()
+        ).resolve().str_with_constraints()
 
     def instance(self) -> TypeInstance:
         return self.schema(*(TypeVar() for _ in range(self.n)))
@@ -175,6 +177,7 @@ class TypeOperator(Type):
         return TypeOperation(self, *(p.instance() for p in params))
 
     def subtype(self, other: TypeOperator, strict: bool = False) -> bool:
+        assert isinstance(other, TypeOperator)
         return ((not strict and self == other) or
             bool(self.supertype and self.supertype.subtype(other)))
 
@@ -189,7 +192,7 @@ class TypeInstance(Type):
     2-ary type operators.
     """
 
-    def strc(self) -> str:
+    def str_with_constraints(self) -> str:
         """
         Like str(), but includes constraints.
         """
@@ -230,19 +233,36 @@ class TypeInstance(Type):
             isinstance(a, TypeOperation) and
             any(b in t for t in a.params))
 
-    def variables(self, distinct: bool = False) -> Iterable[TypeVar]:
+    def variables(self) -> List[TypeVar]:
         """
-        Obtain all distinct type variables currently in the type expression.
+        Obtain all distinct type variables currently in the type expression,
+        including variables that might occur in constraints.
         """
-        return {v.id: v for v in self._variables()}.values() \
-            if distinct else self._variables()
+        # TODO a better way is needed - I'm not sure that id's will remain
+        # distinct in all situations
+        captured = set()
+        result = []
+        for v in self.variables_iter():
+            if v.id not in captured:
+                result.append(v)
+                captured.add(v.id)
+            for c in v.constraints:
+                for v1 in c.variables_iter():
+                    if v1.id not in captured:
+                        result.append(v1)
+                        captured.add(v1.id)
+        return result
 
-    def _variables(self) -> Iterable[TypeVar]:
+    def variables_iter(self) -> Iterable[TypeVar]:
+        """
+        Obtain an iterator of variables in this type instance, excluding
+        variables that might occur in constraints, with possible repetitions.
+        """
         a = self.follow()
         if isinstance(a, TypeVar):
             yield a
         elif isinstance(a, TypeOperation):
-            for v in chain(*(t.variables() for t in a.params)):
+            for v in chain(*(t.variables_iter() for t in a.params)):
                 yield v
 
     def follow(self) -> TypeInstance:
@@ -267,9 +287,13 @@ class TypeInstance(Type):
         else:
             return self
 
-    def subtype(self, other: TypeInstance) -> Optional[bool]:
+    def unifiable(
+            self,
+            other: TypeInstance,
+            subtype: bool = False,
+            accept_wildcard: bool = False) -> Optional[bool]:
         """
-        Return True if self can definitely be a subtype of other, False if it
+        Return True if self can definitely be (a subtype of) other, False if it
         is definitely not, and None if there is not enough information.
         """
         a = self.follow()
@@ -277,16 +301,19 @@ class TypeInstance(Type):
 
         if isinstance(a, TypeOperation) and isinstance(b, TypeOperation):
             if a.basic:
-                return a.operator.subtype(b.operator)
+                if subtype:
+                    return a.operator.subtype(b.operator)
+                else:
+                    return a.operator == b.operator
             elif a.operator != b.operator:
                 return False
             else:
                 result: Optional[bool] = True
                 for v, s, t in zip(a.operator.variance, a.params, b.params):
-                    if v == Variance.CO:
-                        r = s.subtype(t)
-                    elif v == Variance.CONTRA:
-                        r = t.subtype(s)
+                    r = TypeInstance.unifiable(
+                        *((s, t) if v == Variance.CO else (t, s)),
+                        subtype=subtype,
+                        accept_wildcard=accept_wildcard)
                     if r is False:
                         return False
                     elif r is None:
@@ -295,19 +322,23 @@ class TypeInstance(Type):
         elif isinstance(a, TypeOperation) and isinstance(b, TypeVar):
             if (b.upper or b.lower) and not a.basic:
                 return False
-            if b.upper and b.upper.subtype(a.operator, True):
+            if b.upper and b.upper.subtype(a.operator, strict=True):
                 return False
-            if b.wildcard:
+            if accept_wildcard and b.wildcard:
                 return True
         elif isinstance(a, TypeVar) and isinstance(b, TypeOperation):
             if (a.upper or a.lower) and not b.basic:
                 return False
-            if a.lower and b.operator.subtype(a.lower, True):
+            if a.lower and b.operator.subtype(a.lower, strict=True):
                 return False
-            if a.wildcard:
+            if accept_wildcard and a.wildcard:
                 return True
         elif isinstance(a, TypeVar) and isinstance(b, TypeVar):
-            if a.lower and b.upper and a.lower.subtype(b.upper, True):
+            if a == b or (a.wildcard and b.wildcard):
+                return True
+            if accept_wildcard and (a.wildcard or b.wildcard):
+                return True
+            if a.lower and b.upper and a.lower.subtype(b.upper, strict=True):
                 return False
         return None
 
@@ -395,6 +426,8 @@ class TypeOperation(TypeInstance):
             return str(self.operator)
 
     def __eq__(self, other: object) -> bool:
+        if isinstance(other, TypeVar):
+            other = other.follow()
         if isinstance(other, TypeOperation):
             return (self.operator == other.operator and
                 all(s == t for s, t in zip(self.params, other.params)))
@@ -448,6 +481,7 @@ class TypeVar(TypeInstance):
                 constraints = set.union(self.constraints, t.constraints)
                 self.constraints = constraints
                 t.constraints = constraints
+                t.check_constraints()
 
                 t.wildcard = False
 
@@ -459,13 +493,22 @@ class TypeVar(TypeInstance):
                 if t.lower == t.upper and t.lower is not None:
                     t.bind(t.lower())
 
-            elif isinstance(t, TypeOperation) and t.basic:
-                if self.lower and t.operator.subtype(self.lower, True):
-                    raise error.SubtypeMismatch(t, self)
-                if self.upper and self.upper.subtype(t.operator, True):
-                    raise error.SubtypeMismatch(self, t)
+            elif isinstance(t, TypeOperation):
+                if t.basic:
+                    if self.lower and t.operator.subtype(self.lower, strict=True):
+                        raise error.SubtypeMismatch(t, self)
+                    if self.upper and self.upper.subtype(t.operator, strict=True):
+                        raise error.SubtypeMismatch(self, t)
+                else:
+                    constraints = self.constraints.union(*(
+                        v.constraints for v in t.variables()
+                    ))
+                    self.constraints = constraints
+                    for v in t.variables():
+                        v.constraints = constraints
+                        v.check_constraints()
 
-            self.check_constraints(True)
+            self.check_constraints()
 
     def above(self, new: TypeOperator) -> None:
         """
@@ -479,7 +522,7 @@ class TypeVar(TypeInstance):
             pass
         elif lower.subtype(new):  # tighten the lower bound
             self.lower = new
-            self.check_constraints(False)
+            self.check_constraints()
         else:  # fail on bound from other lineage (neither sub- nor supertype)
             raise error.SubtypeMismatch(lower, new)
 
@@ -496,73 +539,108 @@ class TypeVar(TypeInstance):
             pass
         elif new.subtype(upper):
             self.upper = new
-            self.check_constraints(False)
+            self.check_constraints()
         else:
             raise error.SubtypeMismatch(new, upper)
 
-    def check_constraints(self, unify: bool) -> None:
+    def check_constraints(self) -> None:
         self.constraints = set(
-            c for c in self.constraints if not c.fulfilled(unify))
+            c for c in self.constraints if not c.fulfilled())
 
 
 class Constraint(object):
     """
-    A constraint enforces that its subject type is a subtype of one of its
-    object types.
+    A constraint enforces that its reference is a subtype of one of the given
+    alternatives.
     """
 
     def __init__(
             self,
-            subject: TypeInstance,
-            *objects: TypeInstance):
-        self.subject = subject
-        self.objects = list(objects)
+            reference: TypeInstance,
+            *alternatives: TypeInstance):
+        self.reference = reference
+        self.alternatives = list(alternatives)
         self.description = str(self)
-        self.fulfilled()
+        self.skeleton: Optional[TypeInstance] = None
 
         # Inform variables about the constraint present on them
         for v in self.variables():
             assert not v.unified
             v.constraints.add(self)
 
+        self.fulfilled()
+
     def __str__(self) -> str:
-        return f"{self.subject} @ {self.objects}"
+        return f"{self.reference} @ {self.alternatives}"
 
     def set_context(self, context: TypeInstance) -> None:
         self.description = f"{context} | {self}"
 
-    def variables(self) -> Iterable[TypeVar]:
+    def variables(self) -> List[TypeVar]:
+        result = self.reference.variables()
+        for t in self.alternatives:
+            result.extend(t.variables())
+        return result
+
+    def variables_iter(self) -> Iterable[TypeVar]:
         return chain(
-            self.subject.variables(),
-            *(o.variables() for o in self.objects)
+            self.reference.variables_iter(),
+            *(o.variables_iter() for o in self.alternatives)
         )
 
-    def fulfilled(self, unify: bool = True) -> bool:
+    def minimize(self) -> None:
+        """
+        Ensure that constraints don't contain extraneous alternatives:
+        alternatives that are equal to, or more general versions of, other
+        alternatives.
+        """
+        minimized: List[TypeInstance] = []
+        for obj in self.alternatives:
+            add = True
+            for i in range(len(minimized)):
+                if obj.unifiable(minimized[i], subtype=True) is True:
+                    if minimized[i].unifiable(obj, subtype=True) is not True:
+                        minimized[i] = obj
+                    add = False
+            if add:
+                minimized.append(obj)
+        self.alternatives = minimized
+
+    def fulfilled(self) -> bool:
         """
         Check that the constraint has not been violated and raise an error
         otherwise. Additionally, return True if it has been completely
         fulfilled and need not be enforced any longer.
         """
-        compatibility = [self.subject.subtype(t) for t in self.objects]
-        self.objects = [t for t, c in zip(self.objects, compatibility)
+
+        # TODO Minimization is expensive and it is performed on every variable
+        # binding. Make this more efficient?
+        self.minimize()
+
+        compatibility = [
+            self.reference.unifiable(t, subtype=True, accept_wildcard=True)
+            for t in self.alternatives]
+        self.alternatives = [t for t, c in zip(self.alternatives, compatibility)
             if c is not False]
 
-        if len(self.objects) == 0:
+        if len(self.alternatives) == 0:
             raise error.ConstraintViolation(self)
 
         # If there is only one possibility left, we can unify, but *only* with
         # the skeleton: the base types must remain variable, because we don't
         # want to resolve against an overly loose subtype bound.
-        elif len(self.objects) == 1 and unify:
-            self.subject.unify(self.objects[0].skeleton())
+        elif len(self.alternatives) == 1 and not self.skeleton:
+            self.skeleton = self.alternatives[0].skeleton()
+            self.reference.unify(self.skeleton)
 
-        # Fulfillment is achieved if the subject is fully concrete and there is
-        # at least one definitely compatible object
-        return not any(self.subject.variables()) and any(compatibility)
+        # Fulfillment is achieved if the reference is fully concrete and there
+        # is at least one definitely compatible alternative
+        return (not any(self.reference.variables()) and any(compatibility)) or \
+            self.reference in self.alternatives
 
 
 "The special constructor for function types."
-Function = TypeOperator('Function', params=(Variance.CONTRA, Variance.CO))
+Function = TypeOperator('Function', params=[Variance.CONTRA, Variance.CO])
 
 "A wildcard: fresh variable, unrelated to, and matchable with, anything else."
 _ = TypeSchema(lambda: TypeVar(wildcard=True))
